module IPC (
	    toDNAMmodel,toDNAMmodelWithLog, DNAMmodel(..),
	    toDNAMmodelSpec,toDNAMmodelSpecWithLog, DNAMmodelSpec(..)
	   ) where

import Utils (isNothing, fromJust, (?), isRight, fromLeft, selectTerm, isJust, indent, numberLines)
import List (partition, nub, (\\))
import LogResult (LogResult, mapLR, closeLogLR)
import PepaWrap (
		 toPEPAsystem, toPEPAsystemWithLog, toPEPAmodel, toPEPAmodelWithLog, 
		 toGPS, toGPSrtn ,toGPSwithLog, toGPSrtnWithLog,
		 mkPEPAtransList, getPEPApCpts, nextPEPAcpt, fromPEPArate,
		 PEPAsystem, PEPAmodel (..), PEPAprobeDef, PEPAact, PEPAcpt, 
		 PEPAtrans, PEPAdb, PEPAtransList, IPCsystemRtn (..),
		 GenericParallelStructure (..), GPSelem (..)
		)
import PepaTypes (Identifier, Rate, mkLbl, isRateSymbol)

toDNAMmodelSpec :: PEPAprobeDef -> String -> DNAMmodelSpec
toDNAMmodelSpec def = fromGPSrtn . toGPSrtn . (toPEPAsystem def)

toDNAMmodelSpecWithLog :: PEPAprobeDef -> String -> LogResult DNAMmodelSpec
toDNAMmodelSpecWithLog def = fromGPSrtnWithLog . toGPSrtnWithLog . (toPEPAsystemWithLog def)

toDNAMmodel :: String -> DNAMmodel
toDNAMmodel = fromGPS . toGPS . toPEPAmodel

toDNAMmodelWithLog :: String -> LogResult DNAMmodel
toDNAMmodelWithLog = fromGPSwithLog . toGPSwithLog . toPEPAmodelWithLog

fromGPSrtnWithLog :: LogResult (IPCsystemRtn DNAMparStruct) -> LogResult DNAMmodelSpec
fromGPSrtnWithLog  = mapLR "IPC [fromGPSrtn]" fromGPSrtn analyseDNAMmodelSpec

fromGPSwithLog :: LogResult DNAMparStruct -> LogResult DNAMmodel
fromGPSwithLog = mapLR "IPC [fromGPS]" fromGPS analyseDNAMmodel

type DNAMplaceLbl = Identifier
type DNAMplaceTok = Int
type DNAMplace = (DNAMplaceLbl, DNAMplaceTok)
type DNAMtransLbl = Identifier
data DNAMtransCond = DNAMgt DNAMplace | DNAMrtGeq Rate Rate | DNAMnot DNAMtransCond | DNAMand DNAMtransCond DNAMtransCond | DNAMor DNAMtransCond DNAMtransCond | DNAMtt
data DNAMtransAction = DNAMta DNAMplaceLbl DNAMplace -- (p1, (p2, k)) -> { p1 := p2 + k; }
type DNAMtransPriority = Int
data DNAMtransFiring = DNAMmarkov (Maybe [Rate]) deriving Eq
		     -- maintain list of rates so that new duplicate terms can be removed

data DNAMtrans = DNAMt DNAMtransLbl PEPAact [DNAMtransCond] [DNAMtransAction] DNAMtransPriority DNAMtransFiring
type DNAMinit = [DNAMplace]
data DNAMtail = DNAMpassageSpec DNAMtransCond DNAMtransCond DNAMtimeSpec
type DNAMtime = Double
data DNAMtimeSpec = DNAMts (DNAMtime, DNAMtime) DNAMtime
                        -- (\t_start, \t_stop)  \t_step

initDNAMtimeSpec :: DNAMtimeSpec
initDNAMtimeSpec = DNAMts (0.5,25.0) 2.5

data DNAMcpt = DNAMdesc [DNAMplaceLbl] [DNAMtrans]

type DNAMparStruct = GenericParallelStructure DNAMmodel

instance Ord DNAMtransFiring where
    min (DNAMmarkov Nothing) (DNAMmarkov Nothing) = (DNAMmarkov Nothing)
    min (DNAMmarkov Nothing) rt@(DNAMmarkov (Just _)) = rt
    min rt@(DNAMmarkov (Just _)) (DNAMmarkov Nothing) = rt
    min (DNAMmarkov (Just rs1)) (DNAMmarkov (Just rs2)) = (DNAMmarkov $ Just $ nub $ rs1++rs2)

data DNAMmodel = DNAMmod DNAMinit [DNAMtrans]
data DNAMmodelSpec = DNAMms DNAMmodel DNAMtail

instance GPSelem DNAMmodel where
    toGPSelem = getDNAMmodel
    documentGPSelem = analyseDNAMmodel

analyseDNAMmodelSpec :: DNAMmodelSpec -> String
analyseDNAMmodelSpec (DNAMms mod (DNAMpassageSpec sc tc (DNAMts rng st))) = (analyseDNAMmodel mod) 
		      ++ "\nPassage specification:\n"
		      ++ numberLines ["Start condition: " ++ show sc,
				      "Stop condition: " ++ show tc,
				      "Time range: " ++ show rng,
				      "Time step: " ++ show st]

analyseDNAMmodel :: DNAMmodel -> String
analyseDNAMmodel (DNAMmod is ts)
    = concat [show $ length is, " local state(s) found:\n",
	      numberLines (map fst is),"\n",
	      show $ length ts, " local transitions(s) found:\n",
	      numberLines $ map (\(DNAMt l _ _ _ _ _) -> l) ts]
	      

-- Uses DFS on PEPA model to obtain component state space and construct equivalent DNAmaca expression for a given component
-- Int arg used to distinguish components in DNAmaca which have the same name in the PEPA description
-- While its there, getDNAMmodel also sets the start place for each component by inserting a token into the relevent place
getDNAMmodel :: Int -> PEPAmodel -> DNAMmodel
getDNAMmodel i mod@(PEPAmod p cpts) = DNAMmod inits trans where				    
    DNAMdesc lbls trans = getCpt i cpts (mkPEPAtransList mod) [p']
    inits = (p',1):[(q,0) | q <- lbls, q /= p']
    p' = mkDNAMplaceLbl p i
    getCpt :: Int -> PEPAdb -> PEPAtransList -> [DNAMplaceLbl] -> DNAMcpt
    getCpt _ _ [] ls = DNAMdesc ls []
    getCpt i db (ptr@(c1,tr,c2):ptrs) ls = DNAMdesc dlbls (dtr:dtrs) where
         DNAMdesc dlbls dtrs = getCpt i db (nptrs++ptrs) (dlbl'++ls)
	 (dlbl', nptrs) = (elem dlbl ls) ? (([],[]),([dlbl], mkPEPAtransList $ PEPAmod c2 db))
	 (dtr, dlbl) = mapPEPAtrans i c1 (tr,c2) -- Labels transition

instance Eq DNAMtransAction where
    (==) (DNAMta l1 _) (DNAMta l2 _) = (l1==l2)

instance Show DNAMtransCond where
    show (DNAMgt (lbl, n)) = concat [lbl, " > ", show n]
    show (DNAMrtGeq r1 r2) = concat [show r1, " >= ", show r2]
    show (DNAMnot c) = concat ["! ( ",show c," )"]
    show (DNAMand c1 c2) = concat ["(",show c1, ") && (", show c2, ")"]
    show (DNAMor c1 c2) = concat ["(",show c1, ") || (", show c2, ")"]
    show DNAMtt = "IPCTRUE"

instance Show DNAMtransAction where
    show (DNAMta lbl (lbl',tok)) = concat [lbl, " = ", lbl', (tok < 0) ? (" - ", " + "), show $ abs tok, ";"]

instance Show DNAMtransFiring where
    show (DNAMmarkov rts)
	| isNothing rts = "NULL" 
	| otherwise     = show $ foldr1 min $ fromJust rts

instance Show DNAMtrans where
    show (DNAMt lbl act cnds acts pr fr)
	 = showClause ("transition{"++lbl++"}") contents where
	      contents = concat [comment, conditions, actions, priority, rate]
	      comment = concat ["%% PEPA action type { ", act, " }\n"]
	      conditions = concat ["\\condition{ ", foldr1 (\x -> \y -> x++" && "++y) $ map show cnds, " }\n"]
	      actions = (showClause "action" (concat $ map (\x -> "next -> "++(show x)++"\n") acts)) ++ "\n"
	      priority = concat ["\\priority{", show pr, "}\n"]
	      rate = concat ["\\rate{ ", show fr, " }"]
				
instance Show DNAMmodel where
    show (DNAMmod is ts)
	 = concat [comment,"\n\n", body] where 
	      comment = ""
	      preamble = header ++"\n" ++ ((constants == "") ? ("", comment2++"\n"++constants))
	      comment2 = "\n%% Alter the rate constant(s) below as desired:"
	      header = "\\header{#define ipcmin(a,b) (a < b) ? a : b}\n"
	      constants = unlines [ (concat ["\\constant{", show t, "}{1.0}"]) | t <- getDNAMconstants ts ]
	      body = showClause "model" (concat [preamble, "\n\n", vector,"\n\n",initial,"\n\n",transitions]) where
	      vector = showClause "statevector" (showClause "type{short}" (foldr1 (\x -> \y -> x++",\n"++y) $ map (foldr1 (\x -> \y -> x++", "++y)) states))
	      initial = showClause "initial" (init $ concat $ map (\(x,y) -> x++"\t = "++(show y)++";\n") is)
	      transitions = concat $ map (\x -> (show x)++"\n\n") ts
	      states = mkCols 2 $ map fst is
	      mkCols :: Int -> [DNAMplaceLbl] -> [[DNAMplaceLbl]]
	      mkCols m [] = []
	      mkCols m xs = (r:mkCols m rs) where
			    (r, rs) = splitAt m xs


instance Show DNAMtimeSpec where
    show (DNAMts (t1, t2) ts) = unlines ["%% Alter time range figures below as required:",
					 "\\t_start{"++(show t1)++"}", "\\t_stop{"++(show t2)++"}", "\\t_step{"++(show ts)++"}"]

instance Show DNAMtail where
    show (DNAMpassageSpec sc tc ts) = showClause "passage" cnt where
				      cnt = unlines ["\\sourcecondition{ "++show sc++" }",
						     "\\targetcondition{ "++show tc++" }\n"] ++ (show ts)

instance Show DNAMmodelSpec where
    show (DNAMms mod tl) = concat [show mod, "\n\n%% Passage specification\n\n", show tl]


---- PEPA to DNAmaca section ----

getDNAMconstants :: [DNAMtrans] -> [Rate]
getDNAMconstants trs = nub $ filter isRateSymbol $ concat [ fromJust mr | (DNAMt _ _ _ _ _ (DNAMmarkov mr)) <- trs, isJust mr]

showClause :: String -> String -> String
showClause title inset = "\\"++title++" {\n"++(indent inset)++"}"

fromGPSrtn :: IPCsystemRtn DNAMparStruct -> DNAMmodelSpec
fromGPSrtn (IPCsr ps (sl, tl)) = (DNAMms (fromGPS ps) (DNAMpassageSpec (DNAMgt (sl,0)) (DNAMgt (tl,0)) initDNAMtimeSpec))

fromGPS :: DNAMparStruct -> DNAMmodel
fromGPS struct
    | test = error ("IPC.fromGPS: DNAmaca model has unresolved/unsynchronised rates:\n\n" ++ (show lst))
    | otherwise = mod where
         mod = mkDNAMtransUnique $ convertGPS struct
	 (test,lst) = isDNAMmodelOpen mod
	 
-- Guarantees that all transition labels are unique - simply numbers them uniquely
mkDNAMtransUnique :: DNAMmodel -> DNAMmodel
mkDNAMtransUnique (DNAMmod ls ts) = (DNAMmod ls ts') where
		     ts' = [DNAMt ((show i)++"_"++lbl) pa cs as tp tf | (i, DNAMt lbl pa cs as tp tf) <- (zip [1..] ts)]

convertGPS :: DNAMparStruct -> DNAMmodel
convertGPS (SEQ cp) = cp
convertGPS (PAR lhs rhs acts)
    | (length$nub$map fst (ls1++ls2)) < (length ls1)+(length ls2)
	= error ("IPC.convertGPS: repeat states in different DNAM components:\n\n" 
		 ++ (show [l | l <- (map fst ls1), elem l (map fst ls2)]))
    | otherwise = (DNAMmod (ls1++ls2) ts) where
	  DNAMmod ls1 ts1 = convertGPS lhs
	  DNAMmod ls2 ts2 = convertGPS rhs
	  ts = ts1ns++ts2ns++(procDNAMtrans ts1s ts2s acts)
	  (ts1s, ts1ns) = partition (\(DNAMt _ a _ _ _ _) -> elem a acts) ts1
	  (ts2s, ts2ns) = partition (\(DNAMt _ a _ _ _ _) -> elem a acts) ts2

isDNAMmodelOpen :: DNAMmodel -> (Bool, [DNAMtrans])
isDNAMmodelOpen (DNAMmod _ ts) = let ts'= filter (\(DNAMt _ _ _ _ _ (DNAMmarkov r)) -> isNothing r) ts in ((not.null) ts', ts')

procDNAMtrans :: [DNAMtrans] -> [DNAMtrans] -> [PEPAact] -> [DNAMtrans]
procDNAMtrans ts1 ts2 as = [ consDNAMtrans act t1 t2 | act <- as, t1 <- findDNAMact act ts1, t2 <- findDNAMact act ts2 ]

findDNAMact :: PEPAact -> [DNAMtrans] -> [DNAMtrans]
findDNAMact a = filter (\(DNAMt _ a' _ _ _ _) -> (a' == a))


consDNAMtrans :: PEPAact -> DNAMtrans -> DNAMtrans -> DNAMtrans
consDNAMtrans a (DNAMt l1 a1 cs1 as1 p1 f1) (DNAMt l2 a2 cs2 as2 p2 f2)
    | not ((a == a1) && (a1 == a2)) = error "IPC.consDNAMTrans: not equal PEPA action-type transitions"
    | p1 /= p2                      = error "IPC.consDNAMTrans: unequal priorities"
    | otherwise                     = mkTr $ min f1 f2 where
	 mkTr :: DNAMtransFiring -> DNAMtrans
	 mkTr f = (DNAMt (l1++"__"++l2) a (cs1++cs2) (foldr insertDNAMtransAction as1 as2) p1 f)


mapPEPAtrans :: Int -> PEPAcpt -> (PEPAtrans, PEPAcpt) -> (DNAMtrans, DNAMplaceLbl)
mapPEPAtrans i c1 ((a,rt), c2) = (DNAMt lbl a cnds acts pty frg, p2) where
       lbl = p1++"_"++a
       (p1,p2) = (mkDNAMplaceLbl c1 i, mkDNAMplaceLbl c2 i)
       cnds = [(DNAMgt (p1,0))]
       acts = foldr insertDNAMtransAction [] [DNAMta p1 (p1,-1), DNAMta p2 (p2,1)]
       pty = 1
       frg = (DNAMmarkov $ fromPEPArate rt)


insertDNAMtransAction :: DNAMtransAction -> [DNAMtransAction] -> [DNAMtransAction]
insertDNAMtransAction t@(DNAMta lbl (lbl1, i)) ts
    | isNothing fnd = (t:ts)
    | (lbl1 /= lbl2) || (lbl1 /= lbl) = error "IPC.insertDNAMtransAction: strange configuration of transition actions"
    | (i+j) == 0    = ts' -- uncomment to remove {P = P + 0; } style actions
    | otherwise = (DNAMta lbl (lbl, i+j):ts') where
		   fnd = selectTerm (t ==) ts
		   (DNAMta _ (lbl2, j), ts') = fromJust fnd


mkDNAMplaceLbl :: PEPAcpt -> Int -> DNAMplaceLbl
mkDNAMplaceLbl c i = mkLbl (c,i)

main :: String -> (String,String) -> IO ()
main f1 (sa,ta) = do {
	      s <- readFile f1;
	      let 
	        def  = ("TESTER",([sa],[ta]))
--	        lg = fst $ closeLogLR "Test Pepa spec rig" (toDNAMmodelSpecWithLog def s) 
	        lg = toDNAMmodelSpec def s
	      in putStrLn $ show lg
	     }

